/*
 * $Id$
 *
 * The ethz module offers some generic methods for header parsing, 
 * database-driven user authentication and uri classification. 
 * 
 * Copyright (C) 2004 Swiss Federal Institute of Technology (ETH Zurich)
 * Copyright (C) 2004 Marcel Baur <baur@ethworld.ethz.ch>
 *
 * This module is released under the GNU General Public License (GPL).
 * Visit http://www.ethworld.ethz.ch for updated and further information.
 *
 */

/*
 * This file is part of ser, a free SIP server.
 * Copyright (C) 2001-2003 Fhg Fokus
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * History:
 * --------
 *  2004-12-01 Initial version (Marcel Baur)
 */

#include "../../str.h"
#include "../../sr_module.h"
#include "../../db/db.h"
#include "../../action.h"
#include "../../parser/parse_to.h"
#include "../../parser/parse_from.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

MODULE_VERSION
  static int ethz_authorize_register_f (struct sip_msg *msg, char *str,
					char *str2);


static int ethz_is_internal_f (struct sip_msg *msg, char *str, char *str2);

static int ethz_is_external_f (struct sip_msg *msg, char *str, char *str2);

static int ethz_is_inet_f (struct sip_msg *msg, char *str, char *str2);

static int ethz_authorize_headers_f (struct sip_msg *msg, char *str,
				     char *str2);

static int ethz_lookup_alt_f (struct sip_msg *msg, char *str, char *str2);

static int ethz_outbound_int_authorize_f (struct sip_msg *msg, char *str,
					  char *str2);

static int ethz_outbound_ext_authorize_f (struct sip_msg *msg, char *str,
					  char *str2);

static int ethz_outbound_sip_authorize_f (struct sip_msg *msg, char *str,
					  char *str2);

static int ethz_is_b1_f (struct sip_msg *msg, char *str, char *str2);

static int ethz_relay_username_f (struct sip_msg *msg, char *str, char *str2);

static int mod_init (void);

/* ------------------------------------------------------------------ */

static char *db_url = DEFAULT_DB_URL;

static db_con_t *db_con;
static db_func_t *db_f;

static char *FIELD_INETPHONENO = "inetPhoneNo";
static char *FIELD_USERNAME = "username";

/* ------------------------------------------------------------------ */

#define RESULT_YES 1
#define RESULT_DROP_PACKET 0
#define RESULT_NO -1

#define TABLE_SETUP  "ethworld_setup"
#define TABLE_FORWARDS "ethworld_forwards"

#define BOOSTIE_ROUTED_START 20310
#define BOOSTIE_ROUTED_STOP 20314

#define VIMANA_ROUTED_START 20315
#define VIMANA_ROUTED_STOP  20339

/* ------------------------------------------------------------------ */

static cmd_export_t cmds[] = {
  {"ethz_relay_username", ethz_relay_username_f, 0, 0, REQUEST_ROUTE},
  {"ethz_authorize_register", ethz_authorize_register_f, 0, 0, REQUEST_ROUTE},
  {"ethz_is_internal", ethz_is_internal_f, 0, 0, REQUEST_ROUTE},
  {"ethz_is_external", ethz_is_external_f, 0, 0, REQUEST_ROUTE},
  {"ethz_is_inet", ethz_is_inet_f, 0, 0, REQUEST_ROUTE},
  {"ethz_authorize_headers", ethz_authorize_headers_f, 0, 0, REQUEST_ROUTE},
  {"ethz_lookup_alt", ethz_lookup_alt_f, 0, 0, REQUEST_ROUTE},
  {"ethz_outbound_int_authorize", ethz_outbound_int_authorize_f, 0, 0,
   REQUEST_ROUTE},
  {"ethz_outbound_ext_authorize", ethz_outbound_ext_authorize_f, 0, 0,
   REQUEST_ROUTE},
  {"ethz_outbound_sip_authorize", ethz_outbound_sip_authorize_f, 0, 0,
   REQUEST_ROUTE},
  {"ethz_is_b1", ethz_is_b1_f, 0, 0, REQUEST_ROUTE},
  {0, 0, 0, 0, 0}
};

static param_export_t params[] = {
  {"field_inetPhoneNo", STR_PARAM, &FIELD_INETPHONENO},
  {"field_username", STR_PARAM, &FIELD_USERNAME},
  {0, 0, 0}
};

struct module_exports exports = {
  "ethz",
  cmds,
  params,

  mod_init,			/* module initialization function */
  0,				/* response function */
  0,				/* destroy function */
  0,				/* oncancel function */
  0				/* per-child init function */
};

/* ==================================================================== */
/* given a header field 'To: <sip:user@host>',                          */
/* lookup email user@host in LDAP and                                   */
/* return sip url 'sip:32697@test.ethz.ch'                              */

static char *
hole_to (char *body, int size, int crop_domain)
{
  char tmp[255];
  int start, stop;
  int i;

  assert (body);
  if (size > sizeof (tmp))
    {
      size = sizeof (tmp);
    }
  assert (size > 0);

  start = 0;
  stop = 0;
  i = 0;

  if (strncmp (body, "sip:", 4) == 0)
    {
      start = 0;
    }
  else
    {
      while ((body[i] != '<') && (i < size))
	{
	  start++;
	  i++;
	}
      start++;
    }
  assert (start >= 0);

  if (start >= size)
    {
      start = 0;
    }
  assert (start < size);

  while ((body[i] != '>') && (body[i] != 0x0D) && (i < size))
    {
      i++;
    }
  stop = i;

  if (stop >= size)
    {
      stop = size;
    }

  if ((body[start] == 's') && (body[start + 1] == 'i')
      && (body[start + 2] == 'p') && (body[start + 3] = ':'))
    {
      start += 4;
    }

  assert (stop - start > 0);

  strncpy (tmp, &body[start], stop - start);
  tmp[stop - start] = '\0';

  if (crop_domain == 1)
    {
      for (i = 0; i < stop; i++)
	{
	  if (tmp[i] == '@')
	    {
	      tmp[i] = '\0';
	    }
	}
    }

  LOG (L_ERR, "hole_to is '%s'", tmp);
  return strdup (tmp);
}

/* ================================================================== */

static int
mod_init (void)
{
  int res;
  res = bind_dbmod (db_url, db_f);
  if (res < 0)
    {
      LOG (L_ERR, "ethz: Unable to bind to db module (err=%i)\n", res);
      return -1;
    }
  db_con = db_f->init (db_url);

  if (db_con == NULL)
    {
      LOG (L_ERR, "ethz: Error in db_init");
      return RESULT_NO;
    }
  return 0;
}

/* =================================================================== */
/* sip:[2-3][0-9][0-9][0-9][0-9]@test.ethz.ch */

static int
ethz_is_internal_f (struct sip_msg *msg, char *str1, char *str2)
{
  char *to;
  int num, result;
  struct to_body *receiver;

  assert (msg);
  result = RESULT_NO;
  num = 0;

  receiver = get_to (msg);
  LOG (L_ERR, "receiver uri is '%.*s'", receiver->uri.len, receiver->uri.s);

  to = hole_to (receiver->uri.s, receiver->uri.len, 1);
  num = atoi (to);

  if ((num >= 20000) && (num <= 39999))
    {
      result = RESULT_YES;
    }

  if ((num >= BOOSTIE_ROUTED_START) && (num <= BOOSTIE_ROUTED_STOP))
    {
      result = RESULT_NO;
    }

  if ((num >= VIMANA_ROUTED_START) && (num <= VIMANA_ROUTED_STOP))
    {
      /* ETH WORLD */
      result = RESULT_NO;
    }


  LOG (L_ERR, "is_internal(%s): %i", to, result);
  free (to);
  return result;
}

/* =================================================================== */

static int
ethz_is_inet_f (struct sip_msg *msg, char *str, char *str2)
{

  struct to_body *receiver;
  char *to;
  int num, result;

  result = RESULT_NO;
  num = 0;

  receiver = get_to (msg);
  to = hole_to (receiver->uri.s, receiver->uri.len, 1);
  num = atoi (to);

  if ((num >= VIMANA_ROUTED_START) && (num <= VIMANA_ROUTED_STOP))
    {
      /* ETH WORLD */
      result = RESULT_YES;
    }

  free (to);
  LOG (L_ERR, "is_inet: %i", result);
  return result;
}

/* ======================================================================== */
/* uri=~"^sip:0[0-9]*@test.ethz.ch" */

static int
ethz_is_external_f (struct sip_msg *msg, char *str, char *str2)
{
  char *to;
  int num, result;

  result = RESULT_NO;
  num = 0;

  to = hole_to (msg->to->body.s, msg->to->body.len, 1);
  if (to[0] == '0')
    {
      result = RESULT_YES;
    }
  else
    {
      result = RESULT_NO;
    }
  LOG (L_ERR, "is_external: %i", result);
  return result;
}

/* ======================================================================== */
/* AUTHORIZE REGISTER */

static int
ethz_authorize_register_f (struct sip_msg *msg, char *str, char *str2)
{
  /*we registered only 1 param, so we ignore str2 */
  int result, retval;
  db_key_t _k[2];
  db_val_t _v[2];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz;
  char *username;		/* inet phone no */

  db_val_t *values;
  db_row_t *rows;
  int i;

  retval = RESULT_NO;
  username = hole_to (msg->from->body.s, msg->from->body.len, 1);

  /* SELECT */
  _c[0] = "register";
  _c[1] = "use_sip";
  nc = 2;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = FIELD_INETPHONENO;
  VAL_TYPE (_v) = DB_STR;
  VAL_NULL (_v) = 0;
  VAL_STR (_v).s = username;
  VAL_STR (_v).len = strlen (username);
  n = 1;

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);
  if (result < 0)
    {
      LOG (L_ERR, "auth_register: error %i querying db\n", result);
      return RESULT_NO;
    }

  anz = RES_ROW_N (res);

  if (anz == 0)
    {
      /* no entry found */
      db_f->free_result (db_con, res);
      LOG (L_ERR, "No REGISTER entry for '%s'", username);
      return RESULT_NO;
    }
  else
    {
      /* get result */
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);

      for (i = 0; i < anz; i++)
	{
	  if (VAL_TYPE (values) == DB_INT)
	    {
	      if (VAL_INT (values) == 1)
		{
		  retval = RESULT_YES;
		}
	    }
	}
    }

  db_f->free_result (db_con, res);
  return retval;
}

/* ======================================================================== */

static int
sql_select_int_by_string (char *what, const char *from, const char *where_key,
			  const char *where_val)
{
  int result;
  db_key_t _k[2];
  db_val_t _v[2];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz;
  db_row_t *rows;
  db_val_t *values;
  int i, retval;

  /* SELECT */
  _c[0] = what;
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, from);

  /* WHERE */
  _k[0] = where_key;
  VAL_TYPE (_v) = DB_STR;
  VAL_NULL (_v) = 0;
  VAL_STR (_v).s = (char *) where_val;
  VAL_STR (_v).len = strlen (where_val);
  n = 1;

  /* go! */

  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);
  if (result < 0)
    {
      LOG (L_ERR, "sql_select: error %i querying db\n", result);
      return RESULT_NO;
    }

  anz = RES_ROW_N (res);

  if (anz == 0)
    {
      /* no entry found */
      db_f->free_result (db_con, res);
      return RESULT_NO;
    }
  else
    {
      /* get result */

      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);

      for (i = 0; i < anz; i++)
	{
	  if (VAL_TYPE (values) == DB_INT)
	    {
	      retval = VAL_INT (values);
	      return retval;
	    }
	}
    }

  db_f->free_result (db_con, res);
  return RESULT_NO;
}


/* ======================================================================== */

static int
ethz_outbound_int_authorize_f (struct sip_msg *msg, char *str, char *str2)
{
  char *inet;
  int result;

  inet = hole_to (msg->from->body.s, msg->from->body.len, 1);
  result =
    sql_select_int_by_string ("outbound_int", TABLE_SETUP, FIELD_INETPHONENO,
			      inet);
  free (inet);
  if (result == 1)
    {
      result = RESULT_YES;
    }
  else
    {
      result = RESULT_NO;
    }
  LOG (L_ERR, "outbound_int_authorize %i", result);
  return result;
}

/* ======================================================================== */

static int
ethz_outbound_ext_authorize_f (struct sip_msg *msg, char *str, char *str2)
{
  char *inet;
  int result;
  inet = hole_to (msg->from->body.s, msg->from->body.len, 1);
  result =
    sql_select_int_by_string ("outbound_ext", TABLE_SETUP, FIELD_INETPHONENO,
			      inet);
  free (inet);
  if (result == 1)
    {
      result = RESULT_YES;
    }
  else
    {
      result = RESULT_NO;
    }
  LOG (L_ERR, "outbound_ext_authorize %i", result);
  return result;
}

/* ======================================================================== */

static int
ethz_outbound_sip_authorize_f (struct sip_msg *msg, char *str, char *str2)
{
  char *inet;
  int result;

  inet = hole_to (msg->from->body.s, msg->from->body.len, 1);
  result =
    sql_select_int_by_string ("outbound_sip", TABLE_SETUP, FIELD_INETPHONENO,
			      inet);
  free (inet);
  if (result == 1)
    {
      result = RESULT_YES;
    }
  else
    {
      result = RESULT_NO;
    }
  LOG (L_ERR, "outbound_sip_authorize %i", result);
  return result;
}

/* ======================================================================== */

static char *
extract_auth_username (char *digest)
{
  char tmp[255];
  int start, stop, i;

  start = 0;
  i = start;
  LOG (L_ERR, "extract_auth_username: '%s'", digest);
  while ((i < sizeof (tmp)) && (strncmp (&digest[i], "username=\"", 10) != 0))
    {
      start++;
      i++;
    }
  if (i < sizeof (tmp))
    {
      start += strlen ("username=\"");
      i += strlen ("username=\"");
    }
  stop = start;
  while ((i < sizeof (tmp)) && (digest[i] != '"'))
    {
      stop++;
      i++;
    }
  strncpy (tmp, &digest[start], stop - start);
  tmp[stop - start] = '\0';
  return strdup (tmp);
}

/* ======================================================================== */

static int
ethz_authorize_headers_f (struct sip_msg *msg, char *str, char *str2)
{
  db_key_t _k[2];
  db_val_t _v[2];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz, retval, result;
  char *inet, *username;

  username = extract_auth_username (msg->authorization->body.s);
  inet = hole_to (msg->to->body.s, msg->to->body.len, 1);

  retval = RESULT_NO;

  /* SELECT */
  _c[0] = "id";
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = FIELD_INETPHONENO;
  _k[1] = FIELD_USERNAME;
  VAL_TYPE (_v) = DB_STR;
  VAL_NULL (_v) = 0;
  VAL_STR (_v).s = inet;
  VAL_STR (_v).len = strlen (inet);

  VAL_TYPE (&_v[1]) = DB_STR;
  VAL_NULL (&_v[1]) = 0;
  VAL_STR (&_v[1]).s = username;
  VAL_STR (&_v[1]).len = strlen (username);
  n = 2;

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);
  if (result < 0)
    {
      LOG (L_ERR, "authorize_headers: error querying db\n");
      return RESULT_NO;
    }

  anz = RES_ROW_N (res);
  LOG (L_ERR, "authorize_headers: %i results", anz);
  if (anz == 1)
    {
      LOG (L_ERR, "[OK] username='%s', inet='%s'", username, inet);
      retval = RESULT_YES;
    }
  else
    {
      LOG (L_ERR, "[FAIL] username='%s', inet='%s'", username, inet);
      retval = RESULT_NO;
    }

  free (username);
  free (inet);

  return retval;
}

/* ================================================================== */

static void
append_uri (struct sip_msg *msg, const char *uri)
{
  struct action act;
  int ret;

  memset (&act, 0, sizeof (act));
  act.type = APPEND_BRANCH_T;
  act.p1_type = STRING_ST;
  act.p1.string = strdup (uri);
  ret = do_action (&act, msg);
  if (ret < 0)
    {
      LOG (L_ERR, "ERROR: append_uri error");
    }
}


static void
set_uri (struct sip_msg *msg, const char *uri)
{
  struct action act;
  int ret;

  memset (&act, 0, sizeof (act));
  act.type = SET_URI_T;
  act.p1_type = STRING_ST;
  act.p1.string = strdup (uri);
  ret = do_action (&act, msg);
  if (ret < 0)
    {
      LOG (L_ERR, "ERROR: set_uri error");
    }
}



/* ================================================================== */
/* given "baur@ethworld.ethz.ch" as primary email, lookup user name */
/* "baurma" in mysql database */

static char *
ethz_lookup_username_by_email (const char *email)
{
  db_key_t _k[3];
  db_val_t _v[3];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz, result;
  char *tmp;
  db_row_t *rows;
  db_val_t *values;

  /* SELECT */
  _c[0] = FIELD_USERNAME;
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = "email";
  n = 1;

  VAL_TYPE (&_v[0]) = DB_STR;
  VAL_NULL (&_v[0]) = 0;
  VAL_STR (&_v[0]).s = (char *) email;
  VAL_STR (&_v[0]).len = strlen (email);

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);

  if (result < 0)
    {
      LOG (L_ERR, "lookup_username_with_email: no user fond\n");
      return NULL;
    }

  anz = RES_ROW_N (res);
  if (anz != 1)
    {
      return NULL;
    }
  else
    {
      /* got a result, find out which one */
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);
      tmp = VAL_STR (values).s;
      LOG (L_ERR, "user: '%s'\n", tmp);
    }
  return tmp;
}

/* ================================================================ */
/* given "23123" as inet phone no, lookup user name */
/* "baurma" in mysql database */

static char *
ethz_lookup_username_by_inet (const char *inet)
{
  db_key_t _k[3];
  db_val_t _v[3];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz, result;
  char *tmp;
  db_row_t *rows;
  db_val_t *values;

  /* SELECT */
  _c[0] = FIELD_USERNAME;
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = FIELD_INETPHONENO;
  n = 1;

  VAL_TYPE (&_v[0]) = DB_STR;
  VAL_NULL (&_v[0]) = 0;
  VAL_STR (&_v[0]).s = (char *) inet;
  VAL_STR (&_v[0]).len = strlen (inet);

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);

  if (result < 0)
    {
      LOG (L_ERR, "lookup_username_with_email: no user fond\n");
      return NULL;
    }

  anz = RES_ROW_N (res);
  if (anz != 1)
    {
      return NULL;
    }
  else
    {
      /* got a result, find out which one */
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);
      tmp = VAL_STR (values).s;
      LOG (L_ERR, "user: '%s'\n", tmp);
    }
  return tmp;
}


/* ================================================================ */

static char *
ethz_lookup_alt_by_username (const char *username)
{
  db_key_t _k[3];
  db_val_t _v[3];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz, result;
  char *tmp;
  db_row_t *rows;
  db_val_t *values;

  tmp = NULL;

  /* SELECT */
  _c[0] = "sip_url";
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_FORWARDS);

  /* WHERE */
  _k[0] = FIELD_USERNAME;

  VAL_TYPE (&_v[0]) = DB_STR;
  VAL_NULL (&_v[0]) = 0;
  VAL_STR (&_v[0]).s = (char *) username;
  VAL_STR (&_v[0]).len = strlen (username);

  n = 1;

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);
  if (result < 0)
    {
      LOG (L_ERR, "lookup_alt_by_username: user not found\n");
      return NULL;
    }

  anz = RES_ROW_N (res);
  if (anz != 1)
    {
      return NULL;
    }
  else
    {
      /* got a result, find out which one */
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);
      tmp = VAL_STR (values).s;
      LOG (L_ERR, "got: '%s'\n", tmp);
    }

  return tmp;
}

/* ================================================================ */
/* gibt RESULT_YES / RESULT_NO zurueck, je nachdem ob homephone fuer */
/* den Benutzer username gesetzt ist */

static int
ethz_homephone_f (char *username)
{
  int retval;

  retval = sql_select_int_by_string ("homephone", TABLE_SETUP, "username",
				     username);

  if (retval == 1)
    {
      retval = RESULT_YES;
    }
  else
    {
      retval = RESULT_NO;
    }
  return retval;
}

/* ================================================================ */
/* sofern homephone auf 0 gesetzt ist, fuege alt zum destset hinzu */

static int
ethz_lookup_alt_f (struct sip_msg *msg, char *str, char *str2)
{
  int retval;
  char *inet, *email, *username, *tmp;

  email = hole_to (msg->to->body.s, msg->to->body.len, 0);
  LOG (L_ERR, "ethz_lookup_alt for email '%s'", email);

  username = ethz_lookup_username_by_email (email);
  if (username != NULL)
    {
      /* is an email */
      LOG (L_ERR, "username '%s' found for email='%s'", username, email);
    }
  else
    {
      inet = hole_to (msg->to->body.s, msg->to->body.len, 1);	/* WITHOUT EMAIL! */
      LOG (L_ERR, "inet is '%s'", inet);

      username = ethz_lookup_username_by_inet (inet);
      if (username != NULL)
	{
	  /* is an inet */
	  LOG (L_ERR, "username '%s' found for inet='%s'", username, inet);
	}
    }

  retval = RESULT_NO;

  if (username != NULL)
    {

      if (ethz_homephone_f (username) == 1)
	{
	  LOG (L_ERR, "ALT uri is homephone");
	}
      else
	{
	  tmp = ethz_lookup_alt_by_username (username);
	  if (tmp != NULL)
	    {
	      LOG (L_ERR, "ALT uri '%s' found for user='%s'", tmp, username);
	      append_uri (msg, tmp);
	      LOG (L_ERR, "after append_uri");
	      /* no free(tmp) allowed here */
	      retval = RESULT_YES;
	    }
	}
    }
  LOG (L_ERR, "lookup_alt returns %i", retval);
  return retval;
}


/* ================================================================== */

static int
ethz_is_pri_email_and_has_inet (struct sip_msg *msg, const char *email)
{
  db_key_t _k[3];
  db_val_t _v[3];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz, result, retval;
  char *tmp;
  char tmp2[255];
  db_row_t *rows;
  db_val_t *values;

  retval = RESULT_NO;

  /* SELECT */
  _c[0] = FIELD_INETPHONENO;
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = "email";
  n = 1;

  VAL_TYPE (&_v[0]) = DB_STR;
  VAL_NULL (&_v[0]) = 0;
  VAL_STR (&_v[0]).s = (char *) email;
  VAL_STR (&_v[0]).len = strlen (email);

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);

  if (result < 0)
    {
      return RESULT_NO;
    }

  anz = RES_ROW_N (res);
  LOG (L_ERR, "anz is %i", anz);
  if (anz < 1)
    {
      retval = RESULT_NO;
    }
  else
    {
      /* got a result, find out which one */
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);
      tmp = VAL_STR (values).s;
      strcpy (tmp2, "sip:");
      strcat (tmp2, tmp);
      strcat (tmp2, "@test.ethz.ch");
      LOG (L_ERR, "new url is '%s'", tmp2);
      if (strlen (tmp2) > 0)
	{
	  set_uri (msg, tmp2);
	  retval = RESULT_YES;
	}
    }
  LOG (L_ERR, "is_pri_email_or_has_inet %i", retval);
  return retval;
}

/* ============================================================== */

static int
ethz_is_b1_f (struct sip_msg *msg, char *str, char *str2)
{
  char *to;
  int result;
  to = hole_to (msg->to->body.s, msg->to->body.len, 0);
  LOG (L_ERR, "is_b1: email='%s'", to);
  result = ethz_is_pri_email_and_has_inet (msg, to);
  LOG (L_ERR, "is_b1 %i", result);
  return result;
}

/* ============================================================== */


static char *
ethz_lookup_inet_by_username (const char *username)
{
  db_key_t _k[3];
  db_val_t _v[3];
  db_key_t _c[2];
  db_res_t *res;
  int n, nc, anz, result;
  char *tmp;
  db_row_t *rows;
  db_val_t *values;

  tmp = NULL;

  /* SELECT */
  _c[0] = FIELD_INETPHONENO;
  nc = 1;

  /* FROM */
  db_f->use_table (db_con, TABLE_SETUP);

  /* WHERE */
  _k[0] = "username";

  VAL_TYPE (&_v[0]) = DB_STR;
  VAL_NULL (&_v[0]) = 0;
  VAL_STR (&_v[0]).s = (char *) username;
  VAL_STR (&_v[0]).len = strlen (username);

  n = 1;

  /* execute query */
  result = db_f->query (db_con, _k, 0, _v, _c, n, nc, 0, &res);
  if (result < 0)
    {
      LOG (L_ERR, "lookup_inet_by_username: user not found\n");
      return NULL;
    }

  anz = RES_ROW_N (res);
  if (anz != 1)
    {
      return NULL;
    }
  else
    {
      /* got a result, find out which one */
      rows = RES_ROWS (res);
      values = ROW_VALUES (rows);
      tmp = VAL_STR (values).s;

      LOG (L_ERR, "got: '%s'\n", tmp);
    }

  return tmp;
}


static int
ethz_relay_username_f (struct sip_msg *msg, char *str, char *str2)
{
  char *to, *inet;
  char tmp[255];

  to = hole_to (msg->to->body.s, msg->to->body.len, 1);
  if ((to == NULL) || (to == ""))
    {
      return RESULT_NO;
    }

  inet = ethz_lookup_inet_by_username (to);
  if ((inet == NULL) || (inet == ""))
    {
      return RESULT_NO;
    }

  strcpy (tmp, "sip:");
  strcat (tmp, inet);
  strcat (tmp, "@ethz.ch");

  set_uri (msg, tmp);

  return RESULT_YES;
}
